==> kite-1.0b1.tar.gz: kite-1.0b1/modules/System/file.c <==
/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "kite_object.h"
#include "vm/kite_vm.h"

#define CHECK_FILE_OPEN \
    if (this->builtin_data.filevalue == NULL) { \
        kite_object_t *exc = \
            kite_new_exception(thd, "System.exceptions.FileError", \
                               "File is not open"); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
    }

/*****************************************************************************
 * Class constructor
 ****************************************************************************/
KITE_CLASS_METHOD(File_construct) {
    kite_object_t *arg1 = args->builtin_data.listvalue.car;
    kite_object_t *arg2 = 
        args->builtin_data.listvalue.cdr->builtin_data.listvalue.car;
    kite_object_t *file = kite_string_object(thd, arg1);
    kite_object_t *mode = kite_string_object(thd, arg2);

    this->builtin_data.filevalue = 
        fopen(file->builtin_data.stringvalue.string, 
              mode->builtin_data.stringvalue.string);
    if (this->builtin_data.filevalue == NULL) {
        char buf[1024];
        strerror_r(errno, buf, 1024);
        kite_object_t *exc = 
            kite_new_exception(thd, "System.exceptions.FileError", buf);
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
    }

    kite_dereference_object(file);
    kite_dereference_object(mode);
}

/*****************************************************************************
 * Write bytes to file
 ****************************************************************************/
KITE_CLASS_METHOD(File_write) {
    kite_object_t *arg1 = args->builtin_data.listvalue.car;
    kite_object_t *bytes = kite_string_object(thd, arg1);

    CHECK_FILE_OPEN;
    fwrite(bytes->builtin_data.stringvalue.string, 1, 
           bytes->builtin_data.stringvalue.length, 
           this->builtin_data.filevalue);
    kite_dereference_object(bytes);
}

/*****************************************************************************
 * Seek to given position in the file
 ****************************************************************************/
KITE_CLASS_METHOD(File_seek) {
    kite_object_t *arg1 = args->builtin_data.listvalue.car;
    kite_object_t *bytes = kite_int_object(thd, arg1);

    CHECK_FILE_OPEN;
    fseek(this->builtin_data.filevalue, bytes->builtin_data.intvalue, 
          SEEK_SET);
    kite_dereference_object(bytes);
}

/*****************************************************************************
 * Read up to the given number of bytes
 ****************************************************************************/
KITE_CLASS_METHOD(File_read) {
    kite_object_t *arg1 = args->builtin_data.listvalue.car;
    kite_object_t *bytes = kite_int_object(thd, arg1);
    char *buf = calloc(bytes->builtin_data.intvalue + 1, 1);
    int ret;

    CHECK_FILE_OPEN;

    ret = 
        fread(buf, 1, bytes->builtin_data.intvalue, 
              this->builtin_data.filevalue);
    if (ret == 0) {
        free(buf);
        kite_vm_return(thd, kite_new_null(thd));
    }
    kite_object_t *rval = kite_new_string(thd, buf);
    free(buf);
    kite_dereference_object(bytes);
    kite_vm_return(thd, rval);
}

/*****************************************************************************
 * Close given file
 ****************************************************************************/
KITE_CLASS_METHOD(File_close) {
    KITE_NO_ARGS;

    CHECK_FILE_OPEN;
    fclose(this->builtin_data.filevalue);
    this->builtin_data.filevalue = NULL;
}

/*****************************************************************************
 * Read line from file
 ****************************************************************************/
KITE_CLASS_METHOD(File_readline) {
    KITE_NO_ARGS;

    char buf[1024], *ptr = NULL, *rval;
    CHECK_FILE_OPEN;
    do {
        rval = fgets(buf, 1024, this->builtin_data.filevalue);
        if (rval == NULL) break;

        if (ptr == NULL) {
            ptr = malloc(strlen(rval) + 1);
            strcpy(ptr, rval);
        } else {
            ptr = realloc(ptr, strlen(ptr) + strlen(rval) + 1);
            strcat(ptr, rval);
        }
    } while(rval && strlen(buf) >= 1023);

    if (ptr == NULL) {
        kite_vm_return(thd, kite_new_null(thd));
    } else {
        kite_object_t *obj = kite_new_string(thd, ptr);
        free(ptr);
        kite_vm_return(thd, obj);
    }
}

KITE_MODULE_INITIALIZER(file) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *inherits =
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass = kite_new_class(thread, inherits, "file");
    
    /* Add methods common to all classes. */
    kite_add_method(thread, newclass, "__construct__",
        kite_new_method(thread, FUNC_COMPILED, File_construct, 2));
    kite_add_method(thread, newclass, "readline", 
        kite_new_method(thread, FUNC_COMPILED, File_readline, 0));
    kite_add_method(thread, newclass, "read", 
        kite_new_method(thread, FUNC_COMPILED, File_read, 1));
    kite_add_method(thread, newclass, "write", 
        kite_new_method(thread, FUNC_COMPILED, File_write, 1));
    kite_add_method(thread, newclass, "seek", 
        kite_new_method(thread, FUNC_COMPILED, File_seek, 1));
    kite_add_method(thread, newclass, "close", 
        kite_new_method(thread, FUNC_COMPILED, File_close, 0));

    /* Add standard file handles */
    kite_object_t *stdin_object = kite_new_instance(thread, newclass);
    stdin_object->builtin_data.filevalue = stdin;
    kite_object_t *stdout_object = kite_new_instance(thread, newclass);
    stdout_object->builtin_data.filevalue = stdout;
    kite_object_t *stderr_object = kite_new_instance(thread, newclass);
    stderr_object->builtin_data.filevalue = stderr;

    kite_add_property(thread, newclass, "stdin", 1);
    kite_set_property(thread, newclass, "stdin", stdin_object);  
    kite_add_property(thread, newclass, "stdout", 1);
    kite_set_property(thread, newclass, "stdout", stdout_object);  
    kite_add_property(thread, newclass, "stderr", 1);
    kite_set_property(thread, newclass, "stderr", stderr_object);  

    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "file", 1);
    kite_set_property(thread, parent, "file", newclass);
}

==> kite-1.0.0.tar.gz: kite-1.0.0/modules/System/file.c <==
/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "kite_object.h"
#include "vm/kite_vm.h"

#define CHECK_FILE_OPEN \
    if (this->builtin_data.filevalue == NULL) { \
        kite_object_t *exc = \
            kite_new_exception(thd, "System.exceptions.FileError", \
                               "File is not open"); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }

/*****************************************************************************
 * Class constructor
 ****************************************************************************/
KITE_CLASS_METHOD(File_construct) {
    kite_object_t *arg1 = args->builtin_data.listvalue.car;
    kite_object_t *arg2 = 
        args->builtin_data.listvalue.cdr->builtin_data.listvalue.car;
    kite_object_t *file = kite_string_object(thd, arg1);
    kite_object_t *mode = kite_string_object(thd, arg2);
    kite_object_t *exc;

    this->builtin_data.filevalue = 
        fopen(file->builtin_data.stringvalue.string, 
              mode->builtin_data.stringvalue.string);
    if (this->builtin_data.filevalue == NULL) {
        char *buf;
        buf = strerror(errno);
        exc = 
            kite_new_exception(thd, "System.exceptions.FileError", buf);
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }

    kite_dereference_object(file);
    kite_dereference_object(mode);
}

/*****************************************************************************
 * Write bytes to file
 ****************************************************************************/
KITE_CLASS_METHOD(File_write) {
    kite_object_t *arg1 = args->builtin_data.listvalue.car;
    kite_object_t *bytes = kite_string_object(thd, arg1);

    CHECK_FILE_OPEN;
    fwrite(bytes->builtin_data.stringvalue.string, 1, 
           bytes->builtin_data.stringvalue.length, 
           this->builtin_data.filevalue);
    kite_dereference_object(bytes);
}

/*****************************************************************************
 * Seek to given position in the file
 ****************************************************************************/
KITE_CLASS_METHOD(File_seek) {
    kite_object_t *arg1 = args->builtin_data.listvalue.car;
    kite_object_t *bytes = kite_int_object(thd, arg1);

    CHECK_FILE_OPEN;
    fseek(this->builtin_data.filevalue, bytes->builtin_data.intvalue, 
          SEEK_SET);
    kite_dereference_object(bytes);
}

/*****************************************************************************
 * Read up to the given number of bytes
 ****************************************************************************/
KITE_CLASS_METHOD(File_read) {
    kite_object_t *arg1 = args->builtin_data.listvalue.car;
    kite_object_t *bytes = kite_int_object(thd, arg1);
    char *buf = calloc(bytes->builtin_data.intvalue + 1, 1);
    int ret;
    kite_object_t *rval;

    CHECK_FILE_OPEN;

    ret = 
        fread(buf, 1, bytes->builtin_data.intvalue, 
              this->builtin_data.filevalue);
    if (ret == 0) {
        free(buf);
        kite_vm_return(thd, kite_new_null(thd));
    }
    rval = 
        kite_new_string_with_length(
            thd, 
            buf, 
            KITE_GET_INTEGER(bytes)
        );
    free(buf);
    kite_dereference_object(bytes);
    kite_vm_return(thd, rval);
}

/*****************************************************************************
 * Close given file
 ****************************************************************************/
KITE_CLASS_METHOD(File_close) {
    KITE_NO_ARGS;

    CHECK_FILE_OPEN;
    fflush(this->builtin_data.filevalue);
    fclose(this->builtin_data.filevalue);
    this->builtin_data.filevalue = NULL;
}

/*****************************************************************************
 * Emit current position in file.
 ****************************************************************************/
KITE_CLASS_METHOD(File_tell) {
    KITE_NO_ARGS;

    CHECK_FILE_OPEN;
    kite_vm_return(thd, kite_new_integer(thd, ftell(this->builtin_data.filevalue)));
}

/*****************************************************************************
 * Read line from file
 ****************************************************************************/
KITE_CLASS_METHOD(File_readline) {
    char buf[1024], *ptr = NULL, *rval;

    KITE_NO_ARGS;
    CHECK_FILE_OPEN;

    do {
        rval = fgets(buf, 1024, this->builtin_data.filevalue);
        if (rval == NULL) break;

        if (ptr == NULL) {
            ptr = malloc(strlen(rval) + 1);
            strcpy(ptr, rval);
        } else {
            ptr = realloc(ptr, strlen(ptr) + strlen(rval) + 1);
            strcat(ptr, rval);
        }
    } while(rval && strlen(buf) >= 1023);

    if (ptr == NULL) {
        kite_vm_return(thd, kite_new_null(thd));
    } else {
        kite_object_t *obj = kite_new_string(thd, ptr);
        free(ptr);
        kite_vm_return(thd, obj);
    }
}

KITE_MODULE_INITIALIZER(file) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *inherits =
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass = kite_new_class(thread, inherits, "file");
    kite_object_t *stdin_object, *stdout_object, *stderr_object;

    kite_set_docstring(newclass, "Disk file manipulation.");
    
    /* Add methods common to all classes. */
    kite_add_method(thread, newclass, "__construct__",
        kite_new_method_compiled_with_docs(thread, File_construct, 
                                  "Opens a file.", 2,
                                  "name", "Name of file to open.",
                                  "mode", "File mode (r = read, w = write, a = append, + = open for opposite mode too)"));
    kite_add_method(thread, newclass, "readline", 
        kite_new_method_compiled_with_docs(thread, File_readline, 
                                  "Read a single line of text into a string.", 0));
    kite_add_method(thread, newclass, "read", 
        kite_new_method_compiled_with_docs(thread, File_read, 
                                  "Read a stream of bytes into a string.", 1,
                                  "bytes", "Number of bytes to read."));
    kite_add_method(thread, newclass, "write", 
        kite_new_method_compiled_with_docs(thread, File_write, 
                                  "Write string to file.", 1,
                                  "string", "String to write."));
    kite_add_method(thread, newclass, "seek", 
        kite_new_method_compiled_with_docs(thread, File_seek, 
                                  "Move current position in file.", 1,
                                  "pos", "Position in file to move to."));
    kite_add_method(thread, newclass, "tell", 
        kite_new_method_compiled_with_docs(thread, File_tell, 
                                  "Return current position in file.", 0));
    kite_add_method(thread, newclass, "close", 
        kite_new_method_compiled_with_docs(thread, File_close, 
                                  "Closes file and prevents further operations on this object.", 0));

    /* Add standard file handles */
    stdin_object = kite_new_instance(thread, newclass);
    stdin_object->builtin_data.filevalue = stdin;
    stdout_object = kite_new_instance(thread, newclass);
    stdout_object->builtin_data.filevalue = stdout;
    stderr_object = kite_new_instance(thread, newclass);
    stderr_object->builtin_data.filevalue = stderr;

    kite_add_property(thread, newclass, "stdin", 1, "Standard input.");
    kite_set_property(thread, newclass, "stdin", stdin_object);  
    kite_add_property(thread, newclass, "stdout", 1, "Standard output.");
    kite_set_property(thread, newclass, "stdout", stdout_object);  
    kite_add_property(thread, newclass, "stderr", 1, "Standard error.");
    kite_set_property(thread, newclass, "stderr", stderr_object);  

    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "file", 1, NULL);
    kite_set_property(thread, parent, "file", newclass);
}
